#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Workflow interfaces """
from abc import ABC, abstractmethod
from typing import Dict, List, Union

from workspaces.workflow.schema import ProcessingSite, Controller
from workspaces.workflow.schema import Workflow, WorkflowRequest, WorkflowRequestFile


class WorkflowServiceIF(ABC):
    """
    Executes workflows; should be a freestanding service.
    """

    @abstractmethod
    def execute(self, request: WorkflowRequest):
        """
        Execute this workflow against these files.

        :param request:
        :return:               a stream of events from this workflow
        """
        pass

    @abstractmethod
    def attach_file_to_request(self, request: WorkflowRequest, filename: str, content: bytes):
        pass

    @abstractmethod
    def create_workflow_request(self, workflow, argument) -> WorkflowRequest:
        pass


class WorkflowInfoIF(ABC):
    """
    Holds information about workflows.
    """

    @abstractmethod
    def lookup_workflow_definition(self, name: str) -> Workflow:
        """
        Look up the workflow with this name.

        :param name:  Workflow name
        :return:      Workflow instance
        """
        pass

    @abstractmethod
    def lookup_workflow_request(self, request_id: int) -> WorkflowRequest:
        """
        Look up the workflow request with this ID.

        :param request_id: Request ID
        :return Workflow instance
        """
        pass

    @abstractmethod
    def lookup_processing_site(self, site_key: str) -> ProcessingSite:
        """
        Look up the processing site with the given key

        :param site_key: the identifier to look up (NMT, NAASC)
        :return: ProcessingSite instance
        """

    @abstractmethod
    def all_workflows(self) -> List[Workflow]:
        """
        Look up all the workflows that are defined by the system.

        :return all the workflows
        """
        pass

    @abstractmethod
    def all_processing_sites(self) -> List[ProcessingSite]:
        """
        Look up all available processing sites and their attributes

        :return:
        """
        pass

    @abstractmethod
    def lookup_controller(self, controller) -> Controller:
        """
        Look up specified control system
        :param controller: the controller to look up
        :return: Controller
        """
        pass

    @abstractmethod
    def all_controllers(self) -> List[Controller]:
        """
        Look up all linked control systems and their attributes

        :return:
        """
        pass

    @abstractmethod
    def all_external_controllers(self) -> List[Controller]:
        """
        Look up all non-workspaces control systems and their attributes

        :return:
        """
        pass

    @abstractmethod
    def create_workflow_request(self, workflow: Union[str, Workflow], argument: Dict) -> WorkflowRequest:
        """
        Create new workflow request and save to database

        :param workflow: workflow to run
        :param argument: workflow arguments
        :return: new WorkflowRequest
        """
        pass

    def save_request(self, request: WorkflowRequest):
        """
        Save a given request and return an integer identifier for it

        :param request: request to save
        :return:        the request id
        """
        pass

    def save_file(self, request: WorkflowRequest, filename: str, content: bytes) -> WorkflowRequestFile:
        """
        Save a given file for the specified request to the database

        :param request: request to update with new file
        :param filename: filename of new file
        :param content: contents of new file in bytes
        :return:
        """
        pass
